import os
import time

import pytest
from mock import MagicMock

import numpy as np
from numpy.testing import assert_equal

from astropy.io import fits
from astropy.table import Table

try:
    from astropy.samp import SAMPHubServer, SAMPIntegratedClient
except ImportError:
    from astropy.vo.samp import SAMPHubServer, SAMPIntegratedClient

from glue.core import Data, DataCollection, Session
from glue.core.subset import ElementSubsetState

from ..samp_state import SAMPState
from ..samp_client import SAMPClient


class WaitMixin():

    def wait(self, exit_condition):
        for iter in range(50):
            if exit_condition(self):
                break
            time.sleep(0.1)
        else:
            raise Exception("Timed out while waiting for exit condition")


class TestSAMPClientConnectSend(WaitMixin):

    def setup_method(self, method):
        self.state = SAMPState()

        self.session = Session()
        self.data_collection = self.session.data_collection

        self.client = SAMPClient(state=self.state,
                                 session=self.session)

        self.client_ext = SAMPIntegratedClient()

    def teardown_method(self, method):
        self.client_ext.disconnect()
        self.client.stop_samp()

    def test_start_builtin_hub(self):
        assert not self.state.connected
        assert self.state.status == 'Not connected to SAMP Hub'
        self.client.start_samp()
        assert self.state.connected
        assert self.state.status == 'Connected to (glue) SAMP Hub'
        self.client.stop_samp()
        assert not self.state.connected
        assert self.state.status == 'Not connected to SAMP Hub'

    def test_start_external_hub(self):
        assert not self.state.connected
        assert self.state.status == 'Not connected to SAMP Hub'
        self.hub = SAMPHubServer()
        self.hub.start()
        self.client.start_samp()
        assert self.state.connected
        assert self.state.status == 'Connected to SAMP Hub'
        self.client.stop_samp()
        assert not self.state.connected
        assert self.state.status == 'Not connected to SAMP Hub'

    def test_metadata(self):
        self.client.start_samp()
        self.client_ext.connect()
        for client in self.client_ext.get_registered_clients():
            if 'hub' not in client:
                metadata = self.client_ext.get_metadata(client)
                break
        else:
            raise Exception("Client not found")
        assert metadata['samp.name'] == 'glueviz'

    def test_send_data(self):

        receiver = MagicMock()

        def receiver_func(private_key, sender_id, msg_id, mtype, params, extra):
            receiver(private_key, sender_id, msg_id, mtype, params, extra)

        self.client.start_samp()

        self.client_ext.connect()
        self.client_ext.bind_receive_call('*', receiver_func)
        self.client_ext.bind_receive_notification('*', receiver_func)

        data1d = Data(x=[1, 2, 3])
        self.client.send_data(layer=data1d, client=self.client_ext.get_public_id())

        self.wait(lambda x: len(receiver.call_args_list) == 3)

        args, kwargs = receiver.call_args_list[-1]
        assert args[3] == 'table.load.votable'
        assert 'url' in args[4]
        assert 'table-id' in args[4]

        t = Table.read(args[4]['url'], format='votable')
        assert_equal(t['x'], [1, 2, 3])

        receiver.reset_mock()

        data2d = Data(a=[[1, 2], [3, 4]])
        self.client.send_data(layer=data2d)

        self.wait(lambda x: len(receiver.call_args_list) == 1)

        args, kwargs = receiver.call_args_list[-1]
        assert args[3] == 'image.load.fits'
        assert 'url' in args[4]
        assert 'image-id' in args[4]

        hdu = fits.open(args[4]['url'], format='votable')[0]
        assert_equal(hdu.data, [[1, 2], [3, 4]])

        receiver.reset_mock()

        subset1d = data1d.new_subset()
        subset1d.subset_state = ElementSubsetState([0, 2])

        self.client.send_data(layer=subset1d)

        self.wait(lambda x: len(receiver.call_args_list) == 1)

        args, kwargs = receiver.call_args_list[-1]
        assert args[3] == 'table.select.rowList'
        assert 'table-id' in args[4]

        assert_equal(args[4]['row-list'], ['0', '2'])


class TestSAMPClientReceive(WaitMixin):

    def setup_method(self, method):

        self.state = SAMPState()

        self.session = Session()
        self.data_collection = self.session.data_collection

        self.client = SAMPClient(state=self.state,
                                 session=self.session)
        self.client.start_samp()
        self.client.register()

        self.client_ext = SAMPIntegratedClient()
        self.client_ext.connect()

        self.session.edit_subset_mode.edit_subset = []
        self.session.edit_subset_mode.data_collection = self.data_collection

    def teardown_method(self, method):
        self.client_ext.disconnect()
        self.client.unregister()
        self.client.stop_samp()

    @pytest.mark.parametrize('fmt', ['fits', 'votable'])
    def test_receive_table_votable(self, tmpdir, fmt):

        filename = tmpdir.join('test').strpath
        t = Table()
        t['a'] = [1, 2, 3]
        t.write(filename, format=fmt)

        message = {}
        message['samp.mtype'] = 'table.load.' + fmt
        message['samp.params'] = {}
        message['samp.params']['url'] = 'file://' + os.path.abspath(filename)
        message['samp.params']['table-id'] = 'testing'
        message['samp.params']['name'] = 'test_table'

        self.client_ext.call_all('tag', message)

        self.wait(lambda x: len(x.data_collection) == 1)

        assert_equal(self.data_collection[0]['a'], [1, 2, 3])
        assert self.data_collection[0].label == 'test_table'

    def test_receive_image(self, tmpdir):

        filename = tmpdir.join('test').strpath
        image = np.array([[1, 2], [3, 4]])
        fits.writeto(filename, image)

        message = {}
        message['samp.mtype'] = 'image.load.fits'
        message['samp.params'] = {}
        message['samp.params']['url'] = 'file://' + os.path.abspath(filename)
        message['samp.params']['image-id'] = 'testing'
        message['samp.params']['name'] = 'test_image'

        self.client_ext.call_all('tag', message)

        self.wait(lambda x: len(x.data_collection) == 1)

        assert_equal(self.data_collection[0]['PRIMARY'], [[1, 2], [3, 4]])
        assert self.data_collection[0].label == 'test_image'

    def test_receive_highlight(self, tmpdir):

        d = Data(a=[1, 2, 3], label='test')
        d.meta['samp-table-id'] = 'table-123'
        self.data_collection.append(d)

        message = {}
        message['samp.mtype'] = 'table.highlight.row'
        message['samp.params'] = {}
        message['samp.params']['table-id'] = 'table-123'
        message['samp.params']['row'] = 1

        self.client_ext.call_all('tag', message)

        assert len(d.subsets) == 0

        self.state.highlight_is_selection = True

        self.client_ext.call_all('tag', message)

        self.wait(lambda x: len(d.subsets) == 1)

        assert len(d.subsets) == 1
        assert_equal(d.subsets[0].to_mask(), [0, 1, 0])

    def test_receive_row_list(self, tmpdir):

        d = Data(a=[1, 2, 3], label='test')
        d.meta['samp-table-id'] = 'table-123'
        self.data_collection.append(d)

        message = {}
        message['samp.mtype'] = 'table.select.rowList'
        message['samp.params'] = {}
        message['samp.params']['table-id'] = 'table-123'
        message['samp.params']['row-list'] = [0, 2]

        self.client_ext.call_all('tag', message)

        self.wait(lambda x: len(d.subsets) == 1)

        assert len(d.subsets) == 1
        assert_equal(d.subsets[0].to_mask(), [1, 0, 1])

    def test_receive_client_change(self):

        self.wait(lambda x: len(x.state.clients) == 2)

        client = SAMPIntegratedClient()
        client.connect()

        self.wait(lambda x: len(x.state.clients) == 3)

        assert len(self.state.clients) == 3
        client_id = client.get_public_id()
        assert (client_id, client_id) in self.state.clients

        client.disconnect()

        self.wait(lambda x: len(x.state.clients) == 2)

        assert len(self.state.clients) == 2



from __future__ import print_function, division, absolute_import

import os

from qtpy import QtWidgets
from qtpy.QtCore import Signal

from glue.utils import nonpartial
from glue.utils.qt import load_ui
from glue.external.echo.qt import autoconnect_callbacks_to_qt

from glue_samp.samp_client import SAMPClient


class QtSAMPClient(SAMPClient, QtWidgets.QWidget):

    call_received = Signal(object, object, object, object, object, object)
    notification_received = Signal(object, object, object, object, object, object)

    def __init__(self, state=None, session=None, parent=None):

        SAMPClient.__init__(self, state=state, session=session)
        QtWidgets.QWidget.__init__(self, parent=parent)

        self.ui = load_ui('samp_client.ui', self,
                          directory=os.path.dirname(__file__))

        self.state = state

        autoconnect_callbacks_to_qt(self.state, self.ui)

        self.ui.button_start_samp.clicked.connect(nonpartial(self.start_samp))
        self.ui.button_stop_samp.clicked.connect(nonpartial(self.stop_samp))

        self.state.add_callback('connected', self.on_connected_change)
        self.state.add_callback('status', self.on_status_change)

        self.on_connected_change()
        self.on_status_change()

    def on_connected_change(self, *args):

        self.ui.button_start_samp.setEnabled(not self.state.connected)
        self.ui.button_stop_samp.setEnabled(self.state.connected)

        if self.state.connected:

            self.register()

            self.call_received.connect(self.receive_message)
            self.notification_received.connect(self.receive_message)

        else:

            self.unregister()

            try:
                self.call_received.disconnect(self.receive_message)
                self.notification_received.disconnect(self.receive_message)
            except TypeError:
                pass

    def on_status_change(self, *args):
        color = 'green' if self.state.connected else 'red'
        self.ui.text_status.setStyleSheet('color: {0}'.format(color))

    def receive_call(self, private_key, sender_id, msg_id, mtype, params, extra):
        self.call_received.emit(private_key, sender_id, msg_id, mtype, params, extra)
        self.client.reply(msg_id, {"samp.status": "samp.ok", "samp.result": {}})

    def receive_notification(self, private_key, sender_id, msg_id, mtype, params, extra):
        self.notification_received.emit(private_key, sender_id, msg_id, mtype, params, extra)



from __future__ import print_function, division, absolute_import

import os
import uuid
import tempfile
from fnmatch import fnmatch

import numpy as np

try:
    from astropy.samp import SAMPClientError, SAMPHubServer, SAMPIntegratedClient, SAMPHubError
except ImportError:
    from astropy.vo.samp import SAMPClientError, SAMPHubServer, SAMPIntegratedClient, SAMPHubError

from glue import __version__ as glue_version
from glue.core import Data
from glue.logger import logger
from glue.core.data_factories.astropy_table import (astropy_tabular_data_votable,
                                                    astropy_tabular_data_fits)
from glue.core.data_factories.fits import fits_reader
from glue.core.data_exporters.gridded_fits import fits_writer
from glue.core.data_exporters.astropy_table import data_to_astropy_table
from glue.core.subset import ElementSubsetState
from glue.external.echo import delay_callback


__all__ = ['SAMPClient']

ICON_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'glue_samp_icon.png')

MTYPES = ['table.load.votable',
          'table.load.fits',
          'table.highlight.row',
          'table.select.rowList',
          'image.load.fits',
          'samp.hub.event.register',
          'samp.hub.event.unregister']


class SAMPClient(object):

    def __init__(self, state=None, session=None):
        self.state = state
        self.session = session
        self.data_collection = session.data_collection
        self.hub = SAMPHubServer()
        self.client = SAMPIntegratedClient()
        self.state.add_callback('connected', self.on_connected)

    def start_samp(self):
        if not self.client.is_connected:
            try:
                self.client.connect()
            except SAMPHubError:
                try:
                    self.hub.start()
                    self.client.connect()
                except Exception:
                    with delay_callback(self.state, 'connected', 'status'):
                        self.state.connected = False
                        self.state.status = 'Could not connect to Hub'
                else:
                    with delay_callback(self.state, 'connected', 'status'):
                        self.state.connected = True
                        self.state.status = 'Connected to (glue) SAMP Hub'
            except:
                with delay_callback(self.state, 'connected', 'status'):
                    self.state.connected = False
                    self.state.status = 'Could not connect to Hub'
            else:
                with delay_callback(self.state, 'connected', 'status'):
                    self.state.connected = True
                    self.state.status = 'Connected to SAMP Hub'
            self.update_clients()

    def stop_samp(self):
        if self.client.is_connected:
            self.client.disconnect()
        if self.hub.is_running:
            self.hub.stop()
        self.state.connected = False
        self.state.status = 'Not connected to SAMP Hub'
        self.state.clients = []

    def register(self):
        for mtype in MTYPES:
            self.client.bind_receive_call(mtype, self.receive_call)
            self.client.bind_receive_notification(mtype, self.receive_notification)

    def unregister(self):
        try:
            for mtype in MTYPES:
                self.client.unbind_receive_call(mtype)
                self.client.unbind_receive_notification(mtype)
        except (AttributeError, SAMPClientError):
            pass

    def on_connected(self, *args):
        if self.state.connected:
            metadata = {'author.email': 'thomas.robitaille@gmail.com',
                        'author.name': 'Thomas Robitaille',
                        'home.page': 'http://www.glueviz.org',
                        'samp.description.text': 'Multi-dimensional linked data exploration',
                        'samp.documentation.url': 'http://www.glueviz.org',
                        'samp.icon.url': 'file://' + ICON_PATH,
                        'samp.name': 'glueviz',
                        'glue.version': glue_version}
            self.client.declare_metadata(metadata)
            self.update_clients()

    def update_clients(self):
        clients = []
        for client in self.client.get_registered_clients():
            metadata = self.client.get_metadata(client)
            clients.append((client, metadata.get('samp.name', client)))
        self.state.clients = clients

    def send_data(self, layer=None, client=None):

        filename = tempfile.mktemp()

        message = {}
        message["samp.params"] = {}

        if isinstance(layer, Data):

            if layer.ndim == 1:
                table = data_to_astropy_table(layer)
                table.write(filename, format='votable')
                message["samp.mtype"] = "table.load.votable"
                if 'samp-table-id' not in layer.meta:
                    layer.meta['samp-table-id'] = str(uuid.uuid4())
                message["samp.params"]['table-id'] = layer.meta['samp-table-id']
            elif layer.ndim == 2:
                fits_writer(filename, layer)
                message["samp.mtype"] = "image.load.fits"
                if 'samp-image-id' not in layer.meta:
                    layer.meta['samp-image-id'] = str(uuid.uuid4())
                message["samp.params"]['image-id'] = layer.meta['samp-image-id']
            else:
                return

            message["samp.params"]['name'] = layer.label
            message["samp.params"]['url'] = 'file://' + os.path.abspath(filename)

        else:

            message['samp.mtype'] = 'table.select.rowList'

            if layer.ndim == 1:
                message["samp.params"]['table-id'] = layer.data.meta['samp-table-id']
                message["samp.params"]['row-list'] = np.nonzero(layer.to_mask())[0].astype(str).tolist()
            else:
                return

        if client is None:
            self.client.notify_all(message)
        else:
            # Make sure client is subscribed otherwise an exception is raised
            subscriptions = self.client.get_subscriptions(client)
            for mtype in subscriptions:
                if fnmatch(message['samp.mtype'], mtype):
                    self.client.notify(client, message)
                    return
            else:
                return

    def receive_call(self, private_key, sender_id, msg_id, mtype, params, extra):
        self.receive_message(private_key, sender_id, msg_id, mtype, params, extra)
        self.client.reply(msg_id, {"samp.status": "samp.ok", "samp.result": {}})

    def receive_notification(self, private_key, sender_id, msg_id, mtype, params, extra):
        self.receive_message(private_key, sender_id, msg_id, mtype, params, extra)

    def receive_message(self, private_key, sender_id, msg_id, mtype, params, extra):

        logger.info('SAMP: received message - sender_id={0} msg_id={1} mtype={2} '
                    'params={3} extra={4}'.format(sender_id, msg_id, mtype,
                                                  params, extra))

        if mtype.startswith('table.load'):

            if self.table_id_exists(params['table-id']):
                logger.info('SAMP: table with table-id={0} has already '
                            'been read in'.format(params['table-id']))
                return

            logger.info('SAMP: loading table with table-id={0}'.format(params['table-id']))

            if mtype == 'table.load.votable':
                data = astropy_tabular_data_votable(params['url'])
            elif mtype == 'table.load.fits':
                data = astropy_tabular_data_fits(params['url'])
            else:
                logger.info('SAMP: unknown format {0}'.format(mtype.split('.')[-1]))
                return

            if 'name' in params:
                data.label = params['name']

            if 'table-id' in params:
                data.meta['samp-table-id'] = params['table-id']

            self.data_collection.append(data)

        elif mtype.startswith('image.load'):

            if self.image_id_exists(params['image-id']):
                logger.info('SAMP: image with image-id={0} has already '
                            'been read in'.format(params['image-id']))
                return

            logger.info('SAMP: loading image with image-id={0}'.format(params['image-id']))

            if mtype == 'image.load.fits':
                data = fits_reader(params['url'])[0]
            else:
                logger.info('SAMP: unknown format {0}'.format(mtype.split('.')[-1]))
                return

            if 'name' in params:
                data.label = params['name']

            if 'image-id' in params:
                data.meta['samp-image-id'] = params['image-id']

            self.data_collection.append(data)

        elif self.state.highlight_is_selection and mtype == 'table.highlight.row':

            if not self.table_id_exists(params['table-id']):
                return

            data = self.data_from_table_id(params['table-id'])

            subset_state = ElementSubsetState(indices=[params['row']], data=data)

            self.session.edit_subset_mode.update(self.data_collection, subset_state)

        elif mtype == 'table.select.rowList':

            if not self.table_id_exists(params['table-id']):
                return

            data = self.data_from_table_id(params['table-id'])

            rows = np.asarray(params['row-list'], dtype=int)

            subset_state = ElementSubsetState(indices=rows, data=data)

            self.session.edit_subset_mode.update(self.data_collection, subset_state)

        elif mtype == 'samp.hub.event.register' or mtype == 'samp.hub.event.unregister':

            self.update_clients()

    def table_id_exists(self, table_id):
        for data in self.data_collection:
            if data.meta.get('samp-table-id', None) == table_id:
                return True
        else:
            return False

    def data_from_table_id(self, table_id):
        for data in self.data_collection:
            if data.meta.get('samp-table-id', None) == table_id:
                return data
        else:
            raise Exception("Table {0} not found".format(table_id))

    def image_id_exists(self, image_id):
        for data in self.data_collection:
            if data.meta.get('samp-image-id', None) == image_id:
                return True
        else:
            return False

    def data_from_image_id(self, image_id):
        for data in self.data_collection:
            if data.meta.get('samp-image-id', None) == image_id:
                return data
        else:
            raise Exception("image {0} not found".format(image_id))



from __future__ import print_function, division, absolute_import

from glue.external.echo import CallbackProperty
from glue.core.state_objects import State


class SAMPState(State):

    status = CallbackProperty('Not connected to SAMP Hub')
    connected = CallbackProperty(False)
    clients = CallbackProperty([])
    highlight_is_selection = CallbackProperty(False)



# Here we describe what we want the behavior to be under different scenarios.
# First we consider the case of another application sending a SAMP message to
# Glue
#
# table.load.votable, table.load.fits, and image.load.fits: load data if not
# already loaded (based on table-id or image-id)
#
# table.select.rowList: update subset or create new subset based on currently
#                       selected subset in glue (as if external application was
#                       just another viewer)
#
# table.highlight.row: should have an option that if enabled will treat this as
#                      a selection (but by default not?)
#
# Next we define what messages we can emit from glue:
#
# table.load.votable, table.load.fits, and image.load.fits: we should be able to
# send this by control-clicking on datasets in the data collection. Doing this
# could assign a table-id or image-id in the Data.meta.
#
# Table.select.rowList: we should be able to send this by control-clicking on
# subset groups or subsets in the data collection. Only subsets for datasets
# that have a table-id or image-id should be sent.
#
# A full list of 'standard' mtypes can be found here:
#
# http://wiki.ivoa.net/twiki/bin/view/IVOA/SampMTypes#table_load

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = 'undefined'

__all__ = ['__version__', 'setup']


def setup():
    from . import menubar  # noqa



from __future__ import print_function, division, absolute_import

from glue.config import menubar_plugin
from glue.utils.qt import get_qapp

from glue_samp.samp_state import SAMPState
from glue_samp.qt.samp_client import QtSAMPClient
from glue_samp.qt.layer_actions import add_samp_layer_actions

samp_client = None


@menubar_plugin("Open SAMP plugin")
def samp_plugin(session, data_collection):

    global samp_client

    if samp_client is None:

        state = SAMPState()
        samp_client = QtSAMPClient(state=state, data_collection=data_collection)

        # We now add actions to the data collection - however we don't use
        # the @layer_action framework because we want to be able to add
        # sub-menus. TODO: expand @layer_action framework to allow sub-menus.

        add_samp_layer_actions(session, samp_client)

    samp_client.show()
    samp_client.raise_()

    app = get_qapp()
    app.aboutToQuit.connect(samp_client.stop_samp)



from functools import partial

from qtpy import QtWidgets

from glue.app.qt.layer_tree_widget import LayerAction

__all__ = ['add_samp_layer_actions']


class SAMPAction(LayerAction):

    _title = 'SAMP'

    def __init__(self, client, *args, **kwargs):
        super(SAMPAction, self).__init__(*args, **kwargs)
        self.client = client
        menu = SAMPMenu(self)
        self.setMenu(menu)
        self.update_enabled()

    def _can_trigger(self):
        if self.single_selection_data():
            layer = self.selected_layers()[0]
            return layer.ndim == 1 or layer.ndim == 2
        elif self.single_selection_subset():
            layer = self.selected_layers()[0]
            return layer.ndim == 1
        else:
            return False

    def _do_action(self):
        pass

    def _send_to_samp(self, client=None):
        self.client.send_data(layer=self.selected_layers()[0], client=client)


class SAMPMenu(QtWidgets.QMenu):

    def __init__(self, action, *args, **kwargs):
        super(SAMPMenu, self).__init__(*args, **kwargs)
        self.action = action
        self.update_clients([])

    def update_clients(self, clients):
        self.clear()
        if clients:
            self.addAction('Broadcast to all clients', self.action._send_to_samp)
            for client, name in clients:
                self.addAction('Send to {0}'.format(name),
                               partial(self.action._send_to_samp, client=client))
        else:
            action = self.addAction('No connected clients')
            action.setEnabled(False)


def add_samp_layer_actions(session, client):
    layer_tree_widget = session.application._layer_widget
    action = SAMPAction(client, layer_tree_widget)
    menu = action.menu()
    layer_tree_widget.ui.layerTree.addAction(action)
    client.state.add_callback('clients', menu.update_clients)
